"""Support running `ctt` in `pre-commit`."""

import sys
from pathlib import Path

from corallium.log import get_logger
from corallium.shell import capture_shell

logger = get_logger()


def _ls_untracked_dir(base_dir: Path) -> set[Path]:
    """Use git to list all untracked files."""
    cmd = 'git ls-files --directory --exclude-standard --no-empty-dir --others'
    output = capture_shell(cmd=cmd, cwd=base_dir)
    return {base_dir / _d.strip() for _d in output.split('\n') if _d}


def check_for_untracked(base_dir: Path) -> None:
    """Resolve edge case in #3 by raising when pre-commit won't error."""
    if untracked_paths := _ls_untracked_dir(base_dir):
        logger.text('pre-commit error: untracked files must be added', untracked_paths=untracked_paths)
        sys.exit(1)



"""Copier Template Tester.

Based on: https://github.com/copier-org/copier/blob/ccfbc9a923f4228af7ca2bf067493665aa15d07c/tests/helpers.py#L20-L81

"""

import logging
from argparse import ArgumentParser, ArgumentTypeError
from pathlib import Path

from corallium.log import configure_logger, get_logger
from corallium.loggers.plain_printer import plain_printer

from ._config import load_config
from ._pre_commit_support import check_for_untracked
from ._write_output import DEFAULT_TEMPLATE_FILE_NAME, read_copier_template, write_output

configure_logger(log_level=logging.INFO, logger=plain_printer)
logger = get_logger()


def run(*, base_dir: Path | None = None, check_untracked: bool = False) -> None:
    """Entry point."""
    base_dir = base_dir or Path.cwd()
    try:
        read_copier_template(base_dir=base_dir)
    except FileNotFoundError:
        message = f"Please add a '{DEFAULT_TEMPLATE_FILE_NAME}' file to '{base_dir}'"
        logger.warning(message)
        return

    logger.text(f'Starting Copier Template Tester for {base_dir}')
    logger.text('\tNote: If files were modified, pre-commit will report a failure.')
    logger.text('')
    config = load_config(base_dir)
    defaults = config.get('defaults', {})

    input_path = base_dir
    paths = set()
    for key, data in config['output'].items():
        output_path = base_dir / key
        paths.add(output_path)
        logger.text(f'Using `copier` to create: {key}')
        write_output(src_path=input_path, dst_path=base_dir / output_path, data=defaults | data)

    if check_untracked:  # pragma: no cover
        check_for_untracked(base_dir)


def run_cli() -> None:  # pragma: no cover
    """Accept CLI configuration for running ctt."""
    def dir_path(pth: str | None) -> Path:
        if pth and Path(pth).is_dir():
            return Path(pth).resolve()
        msg = f'Expected a path to a directory. Received: `{pth}`'
        raise ArgumentTypeError(msg)

    cli = ArgumentParser()
    cli.add_argument(
        '-b',
        '--base-dir',
        help='Specify the path to the directory that contains the configuration file',
        type=dir_path)
    cli.add_argument('--check-untracked', help='Only used for pre-commit', action='store_true')

    args = cli.parse_args()
    run(base_dir=args.base_dir, check_untracked=args.check_untracked)



"""Conditionally configure runtime typechecking."""

from contextlib import suppress
from datetime import datetime, timezone
from enum import Enum
from os import getenv
from warnings import filterwarnings

from typing_extensions import Self

NAME = 'copier_template_tester'.upper()
"""Package name to allow more targeted usage."""


class _RuntimeTypeCheckingModes(Enum):
    """Supported global runtime type checking modes."""

    ERROR = 'ERROR'
    WARNING = 'WARNING'
    OFF = None

    @classmethod
    def from_environment(cls) -> Self:  # pragma: no cover
        """Return the configured mode.

        Raises:
            ValueError: if environment variable is configured incorrectly

        """
        rtc_mode = getenv('RUNTIME_TYPE_CHECKING_MODE') or getenv(f'RUNTIME_TYPE_CHECKING_MODE_{NAME}') or None
        try:
            return cls(rtc_mode)
        except ValueError:
            modes = [_e.value for _e in cls]
            msg = f"'RUNTIME_TYPE_CHECKING_MODE={rtc_mode}' is not from {modes}"
            raise ValueError(msg) from None


def configure_runtime_type_checking_mode() -> None:  # pragma: no cover
    """Optionally configure runtime type checking mode globally."""
    rtc_mode = _RuntimeTypeCheckingModes.from_environment()

    if rtc_mode is not _RuntimeTypeCheckingModes.OFF:
        with suppress(ImportError, ModuleNotFoundError):
            # Requires beartype >=0.15.0 and Python >= 3.8
            from beartype import BeartypeConf  # noqa: PLC0415
            from beartype.claw import beartype_this_package  # noqa: PLC0415
            from beartype.roar import BeartypeClawDecorWarning  # noqa: PLC0415

            beartype_this_package(
                conf=BeartypeConf(
                    warning_cls_on_decorator_exception=(
                        None if rtc_mode is _RuntimeTypeCheckingModes.ERROR else BeartypeClawDecorWarning
                    ),
                ),
            )


_PEP585_DATE = 2025
if datetime.now(tz=timezone.utc).year <= _PEP585_DATE:  # pragma: no cover
    with suppress(ImportError, ModuleNotFoundError):
        from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

        filterwarnings(
            'ignore',
            category=BeartypeDecorHintPep585DeprecationWarning,
        )



"""copier_template_tester."""

from ._runtime_type_check_setup import configure_runtime_type_checking_mode

__version__ = '2.1.4'
__pkg_name__ = 'copier_template_tester'

configure_runtime_type_checking_mode()


# == Above code must always be first ==



"""Template Directory Writer."""

import re
import shutil
import stat
import sys
from contextlib import contextmanager, suppress
from functools import lru_cache
from pathlib import Path

import copier
from copier.template import load_template_config
from corallium.file_helpers import read_lines
from corallium.log import get_logger
from corallium.shell import capture_shell

logger = get_logger()

DEFAULT_TEMPLATE_FILE_NAME = 'copier.yaml'
"""Default answer file name; however, `copier.yml` is also supported through `read_copier_template`."""


DEFAULT_ANSWER_FILE_NAME = '.copier-answers.yml'
"""Default answer file name.

https://github.com/copier-org/copier/blob/7f05baf4f004a4876fb6158e1c532b28290146a4/copier/subproject.py#L39

"""


@lru_cache(maxsize=1)
def read_copier_template(base_dir: Path) -> dict:  # type: ignore[type-arg]
    """Locate the copier file regardless of variation and return the content.

    https://github.com/copier-org/copier/blob/5827d6a6fc6592e64c983bc52a254471ecff7531/docs/creating.md?plain=1#L13-L14

    """
    copier_path = base_dir / DEFAULT_TEMPLATE_FILE_NAME
    if not copier_path.is_file():
        copier_path = copier_path.with_suffix('.yml')
    if not copier_path.is_file():  # pragma: no cover
        msg = f"Can't find the copier template file. Expected: {copier_path} (or .yaml)"
        raise FileNotFoundError(msg)

    return load_template_config(conf_path=copier_path)


@lru_cache(maxsize=1)
def _find_answers_file(*, src_path: Path, dst_path: Path) -> Path:
    """Locate the copier answers file based on the copier template."""
    copier_config = read_copier_template(src_path)
    answers_filename = copier_config.get('_answers_file') or DEFAULT_ANSWER_FILE_NAME
    if '{{' in answers_filename:
        # If the filename is created from the template, just grab the first match
        search_name = re.sub(r'{{[^}]+}}', '*', answers_filename)
        matches = [*dst_path.glob(search_name)]
        if len(matches) != 1:  # pragma: no cover
            msg = f"Can't find just one copier answers file matching {dst_path / search_name}. Found: {matches}"
            raise ValueError(msg)
        return matches[0]
    return dst_path / answers_filename  # pragma: no cover


@lru_cache(maxsize=3)
def _resolve_git_root_dir(base_dir: Path) -> Path:
    """Use git to list all untracked files."""
    cmd = 'git rev-parse --show-toplevel'
    output = capture_shell(cmd=cmd, cwd=base_dir)
    return Path(output.strip())


def _stabilize(line: str, answers_path: Path) -> str:
    # Convert _src_path to a deterministic relative path
    if line.startswith('_src_path'):
        logger.info('Replacing with deterministic value', line=line)
        raw_path = Path(line.split('_src_path:')[-1].strip())
        ans_dir = answers_path.parent
        if ans_dir.is_relative_to(raw_path):
            count_rel = len(ans_dir.relative_to(raw_path).parts)
            rel_path = '/'.join([*(['..'] * count_rel), raw_path.name])
            return f'_src_path: {rel_path}'
        return line
    # Create a stable tag for '_commit' that copier will still utilize
    if line.startswith('_commit'):
        logger.info('Replacing with deterministic value', line=line)
        return '_commit: HEAD'
    return line


def _stabilize_answers_file(*, src_path: Path, dst_path: Path) -> None:
    """Ensure that the answers file is deterministic."""
    answers_path = _find_answers_file(src_path=src_path, dst_path=dst_path)
    lines = (_stabilize(_l, answers_path) for _l in read_lines(answers_path) if _l.strip())
    answers_path.write_text('\n'.join(lines) + '\n')


@contextmanager
# PLANNED: In python 3.10, there is a Beartype error for this return annotation:
#   -> Generator[None, None, None]
def _output_dir(*, src_path: Path, dst_path: Path):  # noqa: ANN202
    """Manage the output directory and handle templates that cannot be updated (i.e. not answers file).

    Addresses: <https://github.com/KyleKing/copier-template-tester/issues/24>

    """
    template_name = '{{ _copier_conf.answers_file }}.jinja'
    has_answers_template = any(src_path.rglob(template_name))

    dst_path.mkdir(parents=True, exist_ok=True)
    yield

    if has_answers_template:
        # Reduce variability in the output
        try:
            _stabilize_answers_file(src_path=src_path, dst_path=dst_path)
        except FileNotFoundError as exc:  # pragma: no cover
            logger.warning(str(exc))
            raise
    else:
        with suppress(FileNotFoundError):
            answers_path = _find_answers_file(src_path=src_path, dst_path=dst_path)
            answers_path.unlink()


def _remove_readonly(func, path: str, _excinfo) -> None:  # noqa: ANN001
    """Clear the readonly bit for `shutil.rmtree(..., onexc=_remove_readonly)`.

    Adapted from: https://docs.python.org/3/library/shutil.html#rmtree-example

    Resolves: https://github.com/KyleKing/copier-template-tester/issues/34

    The first parameter, function, is the function which raised the exception; it depends on the platform and
    implementation. The second parameter, path, will be the path name passed to function. The third parameter,
    excinfo, is the exception that was raised. Exceptions raised by onexc will not be caught.

    """
    Path.chmod(Path(path), stat.S_IWRITE)
    func(path)


def write_output(*, src_path: Path, dst_path: Path, data: dict[str, bool | int | float | str | None], **kwargs) -> None:
    """Copy the specified directory to the target location with provided data.

    kwargs documentation: https://github.com/copier-org/copier/blob/103828b59fd9eb671b5ffa909004d1577742300b/copier/main.py#L86-L173

    """
    with _output_dir(src_path=src_path, dst_path=dst_path):
        kwargs.setdefault('cleanup_on_error', False)
        kwargs.setdefault('data', data or {})
        kwargs.setdefault('defaults', True)
        kwargs.setdefault('exclude', ['.ctt', 'ctt.toml'])
        kwargs.setdefault('overwrite', True)
        kwargs.setdefault('quiet', False)
        kwargs.setdefault('unsafe', True)
        kwargs.setdefault('vcs_ref', 'HEAD')
        copier.run_copy(str(src_path), dst_path, **kwargs)

        # Remove any .git directory created by copier script
        git_path = dst_path / '.git'
        if git_path.is_dir():  # pragma: no cover
            logger.info('Removing git created by copier', git_path=git_path)
            if sys.version_info >= (3, 12, 0):
                shutil.rmtree(git_path, onexc=_remove_readonly)  # type: ignore[call-arg]
            else:
                shutil.rmtree(git_path, onerror=_remove_readonly)



"""CTT config."""

from pathlib import Path

from corallium.log import get_logger
from corallium.tomllib import tomllib

logger = get_logger()


def _validate_config(config: dict) -> None:  # type: ignore[type-arg]
    if 'defaults' not in config:
        logger.text('Warning: You probably want a section: [defaults]')
    if not config.get('output'):
        raise RuntimeError('CTT expected headers like: [output."<something>"]')


def load_config(base_dir: Path) -> dict:  # type: ignore[type-arg]
    """Read the ctt config from `CWD`."""
    cfg_path = base_dir / 'ctt.toml'
    if not cfg_path.is_file():  # pragma: no cover
        msg = f'No configuration file found. Expected: {cfg_path.absolute()}'
        raise ValueError(msg)
    config: dict = tomllib.loads(cfg_path.read_text())  # type: ignore[type-arg]
    _validate_config(config)
    return config
